@@ -18,16 +18,19 @@ Tests for volume transfer code.
 """
 
 import mock
-
 from oslo_serialization import jsonutils
+from six.moves import http_client
 import webob
 
 from cinder.api.contrib import volume_transfer
 from cinder import context
 from cinder import db
 from cinder import exception
+from cinder.objects import fields
+from cinder import quota
 from cinder import test
 from cinder.tests.unit.api import fakes
+from cinder.tests.unit.api.v2 import fakes as v2_fakes
 from cinder.tests.unit import fake_constants as fake
 import cinder.transfer
 
@@ -54,7 +57,8 @@ class VolumeTransferAPITestCase(test.TestCase):
                        display_description='this is a test volume',
                        status='available',
                        size=1,
-                       project_id=fake.PROJECT_ID):
+                       project_id=fake.PROJECT_ID,
+                       attach_status=fields.VolumeAttachStatus.DETACHED):
         """Create a volume object."""
         vol = {}
         vol['host'] = 'fake_host'
@@ -64,8 +68,9 @@ class VolumeTransferAPITestCase(test.TestCase):
         vol['status'] = status
         vol['display_name'] = display_name
         vol['display_description'] = display_description
-        vol['attach_status'] = status
+        vol['attach_status'] = attach_status
         vol['availability_zone'] = 'fake_zone'
+        vol['volume_type_id'] = fake.VOLUME_TYPE_ID
         return db.volume_create(context.get_admin_context(), vol)['id']
 
     def test_show_transfer(self):
@@ -78,7 +83,7 @@ class VolumeTransferAPITestCase(test.TestCase):
         res = req.get_response(fakes.wsgi_app(
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
-        self.assertEqual(200, res.status_int)
+        self.assertEqual(http_client.OK, res.status_int)
         self.assertEqual('test_transfer', res_dict['transfer']['name'])
         self.assertEqual(transfer['id'], res_dict['transfer']['id'])
         self.assertEqual(volume_id, res_dict['transfer']['volume_id'])
@@ -95,8 +100,9 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(404, res.status_int)
-        self.assertEqual(404, res_dict['itemNotFound']['code'])
+        self.assertEqual(http_client.NOT_FOUND, res.status_int)
+        self.assertEqual(http_client.NOT_FOUND,
+                         res_dict['itemNotFound']['code'])
         self.assertEqual('Transfer %s could not be found.' %
                          fake.WILL_NOT_BE_FOUND_ID,
                          res_dict['itemNotFound']['message'])
@@ -115,7 +121,7 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(200, res.status_int)
+        self.assertEqual(http_client.OK, res.status_int)
         self.assertEqual(4, len(res_dict['transfers'][0]))
         self.assertEqual(transfer1['id'], res_dict['transfers'][0]['id'])
         self.assertEqual('test_transfer', res_dict['transfers'][0]['name'])
@@ -142,7 +148,7 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(200, res.status_int)
+        self.assertEqual(http_client.OK, res.status_int)
         self.assertEqual(5, len(res_dict['transfers'][0]))
         self.assertEqual('test_transfer',
                          res_dict['transfers'][0]['name'])
@@ -182,9 +188,7 @@ class VolumeTransferAPITestCase(test.TestCase):
         db.transfer_destroy(context.get_admin_context(), transfer1['id'])
         db.volume_destroy(context.get_admin_context(), volume_id_1)
 
-    @mock.patch(
-        'cinder.api.openstack.wsgi.Controller.validate_string_length')
-    def test_create_transfer_json(self, mock_validate):
+    def test_create_transfer_json(self):
         volume_id = self._create_volume(status='available', size=5)
         body = {"transfer": {"name": "transfer1",
                              "volume_id": volume_id}}
@@ -199,13 +203,12 @@ class VolumeTransferAPITestCase(test.TestCase):
 
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(202, res.status_int)
+        self.assertEqual(http_client.ACCEPTED, res.status_int)
         self.assertIn('id', res_dict['transfer'])
         self.assertIn('auth_key', res_dict['transfer'])
         self.assertIn('created_at', res_dict['transfer'])
         self.assertIn('name', res_dict['transfer'])
         self.assertIn('volume_id', res_dict['transfer'])
-        self.assertTrue(mock_validate.called)
 
         db.volume_destroy(context.get_admin_context(), volume_id)
 
@@ -220,11 +223,9 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(400, res.status_int)
-        self.assertEqual(400, res_dict['badRequest']['code'])
-        self.assertEqual("Missing required element 'transfer' in "
-                         "request body.",
-                         res_dict['badRequest']['message'])
+        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
+        self.assertEqual(http_client.BAD_REQUEST,
+                         res_dict['badRequest']['code'])
 
     def test_create_transfer_with_body_KeyError(self):
         body = {"transfer": {"name": "transfer1"}}
@@ -237,12 +238,11 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(400, res.status_int)
-        self.assertEqual(400, res_dict['badRequest']['code'])
-        self.assertEqual('Incorrect request body format',
-                         res_dict['badRequest']['message'])
+        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
+        self.assertEqual(http_client.BAD_REQUEST,
+                         res_dict['badRequest']['code'])
 
-    def test_create_transfer_with_VolumeNotFound(self):
+    def test_create_transfer_with_invalid_volume_id_value(self):
         body = {"transfer": {"name": "transfer1",
                              "volume_id": 1234}}
 
@@ -255,10 +255,9 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(404, res.status_int)
-        self.assertEqual(404, res_dict['itemNotFound']['code'])
-        self.assertEqual('Volume 1234 could not be found.',
-                         res_dict['itemNotFound']['message'])
+        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
+        self.assertEqual(http_client.BAD_REQUEST,
+                         res_dict['badRequest']['code'])
 
     def test_create_transfer_with_InvalidVolume(self):
         volume_id = self._create_volume(status='attached')
@@ -273,13 +272,34 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(400, res.status_int)
-        self.assertEqual(400, res_dict['badRequest']['code'])
+        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
+        self.assertEqual(http_client.BAD_REQUEST,
+                         res_dict['badRequest']['code'])
         self.assertEqual('Invalid volume: status must be available',
                          res_dict['badRequest']['message'])
 
         db.volume_destroy(context.get_admin_context(), volume_id)
 
+    def test_create_transfer_with_leading_trailing_spaces_for_name(self):
+        volume_id = self._create_volume(status='available', size=5)
+        body = {"transfer": {"name": "    transfer1   ",
+                             "volume_id": volume_id}}
+
+        req = webob.Request.blank('/v2/%s/os-volume-transfer' %
+                                  fake.PROJECT_ID)
+        req.method = 'POST'
+        req.headers['Content-Type'] = 'application/json'
+        req.body = jsonutils.dump_as_bytes(body)
+        res = req.get_response(fakes.wsgi_app(
+            fake_auth_context=self.user_ctxt))
+
+        res_dict = jsonutils.loads(res.body)
+
+        self.assertEqual(http_client.ACCEPTED, res.status_int)
+        self.assertEqual(body['transfer']['name'].strip(),
+                         res_dict['transfer']['name'])
+        db.volume_destroy(context.get_admin_context(), volume_id)
+
     def test_delete_transfer_awaiting_transfer(self):
         volume_id = self._create_volume()
         transfer = self._create_transfer(volume_id)
@@ -290,7 +310,7 @@ class VolumeTransferAPITestCase(test.TestCase):
         res = req.get_response(fakes.wsgi_app(
             fake_auth_context=self.user_ctxt))
 
-        self.assertEqual(202, res.status_int)
+        self.assertEqual(http_client.ACCEPTED, res.status_int)
 
         # verify transfer has been deleted
         req = webob.Request.blank('/v2/%s/os-volume-transfer/%s' % (
@@ -301,8 +321,9 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(404, res.status_int)
-        self.assertEqual(404, res_dict['itemNotFound']['code'])
+        self.assertEqual(http_client.NOT_FOUND, res.status_int)
+        self.assertEqual(http_client.NOT_FOUND,
+                         res_dict['itemNotFound']['code'])
         self.assertEqual('Transfer %s could not be found.' % transfer['id'],
                          res_dict['itemNotFound']['message'])
         self.assertEqual(db.volume_get(context.get_admin_context(),
@@ -319,19 +340,21 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(404, res.status_int)
-        self.assertEqual(404, res_dict['itemNotFound']['code'])
+        self.assertEqual(http_client.NOT_FOUND, res.status_int)
+        self.assertEqual(http_client.NOT_FOUND,
+                         res_dict['itemNotFound']['code'])
         self.assertEqual('Transfer %s could not be found.' %
                          fake.WILL_NOT_BE_FOUND_ID,
                          res_dict['itemNotFound']['message'])
 
-    def test_accept_transfer_volume_id_specified_json(self):
+    @mock.patch.object(quota.QUOTAS, 'reserve')
+    @mock.patch.object(db, 'volume_type_get', v2_fakes.fake_volume_type_get)
+    def test_accept_transfer_volume_id_specified_json(self, type_get):
         volume_id = self._create_volume()
         transfer = self._create_transfer(volume_id)
 
         svc = self.start_service('volume', host='fake_host')
-        body = {"accept": {"id": transfer['id'],
-                           "auth_key": transfer['auth_key']}}
+        body = {"accept": {"auth_key": transfer['auth_key']}}
         req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                   fake.PROJECT_ID, transfer['id']))
         req.method = 'POST'
@@ -341,7 +364,7 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(202, res.status_int)
+        self.assertEqual(http_client.ACCEPTED, res.status_int)
         self.assertEqual(transfer['id'], res_dict['transfer']['id'])
         self.assertEqual(volume_id, res_dict['transfer']['volume_id'])
         # cleanup
@@ -361,11 +384,11 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(400, res.status_int)
-        self.assertEqual(400, res_dict['badRequest']['code'])
-        self.assertEqual("Missing required element 'accept' in request body.",
-                         res_dict['badRequest']['message'])
+        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
+        self.assertEqual(http_client.BAD_REQUEST,
+                         res_dict['badRequest']['code'])
 
+        db.transfer_destroy(context.get_admin_context(), transfer['id'])
         db.volume_destroy(context.get_admin_context(), volume_id)
 
     def test_accept_transfer_with_body_KeyError(self):
@@ -384,17 +407,18 @@ class VolumeTransferAPITestCase(test.TestCase):
 
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(400, res.status_int)
-        self.assertEqual(400, res_dict['badRequest']['code'])
-        self.assertEqual("Missing required element 'accept' in request body.",
-                         res_dict['badRequest']['message'])
+        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
+        self.assertEqual(http_client.BAD_REQUEST,
+                         res_dict['badRequest']['code'])
+
+        db.transfer_destroy(context.get_admin_context(), transfer['id'])
+        db.volume_destroy(context.get_admin_context(), volume_id)
 
     def test_accept_transfer_invalid_id_auth_key(self):
         volume_id = self._create_volume()
         transfer = self._create_transfer(volume_id)
 
-        body = {"accept": {"id": transfer['id'],
-                           "auth_key": 1}}
+        body = {"accept": {"auth_key": 1}}
         req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                   fake.PROJECT_ID, transfer['id']))
         req.method = 'POST'
@@ -404,8 +428,9 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(400, res.status_int)
-        self.assertEqual(400, res_dict['badRequest']['code'])
+        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
+        self.assertEqual(http_client.BAD_REQUEST,
+                         res_dict['badRequest']['code'])
         self.assertEqual(res_dict['badRequest']['message'],
                          'Invalid auth key: Attempt to transfer %s with '
                          'invalid auth key.' % transfer['id'])
@@ -417,8 +442,7 @@ class VolumeTransferAPITestCase(test.TestCase):
         volume_id = self._create_volume()
         transfer = self._create_transfer(volume_id)
 
-        body = {"accept": {"id": transfer['id'],
-                           "auth_key": 1}}
+        body = {"accept": {"auth_key": 1}}
         req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
             fake.PROJECT_ID, fake.WILL_NOT_BE_FOUND_ID))
         req.method = 'POST'
@@ -428,8 +452,9 @@ class VolumeTransferAPITestCase(test.TestCase):
             fake_auth_context=self.user_ctxt))
         res_dict = jsonutils.loads(res.body)
 
-        self.assertEqual(404, res.status_int)
-        self.assertEqual(404, res_dict['itemNotFound']['code'])
+        self.assertEqual(http_client.NOT_FOUND, res.status_int)
+        self.assertEqual(http_client.NOT_FOUND,
+                         res_dict['itemNotFound']['code'])
         self.assertEqual('Transfer %s could not be found.' %
                          fake.WILL_NOT_BE_FOUND_ID,
                          res_dict['itemNotFound']['message'])
@@ -453,8 +478,7 @@ class VolumeTransferAPITestCase(test.TestCase):
         volume_id = self._create_volume()
         transfer = self._create_transfer(volume_id)
 
-        body = {"accept": {"id": transfer['id'],
-                           "auth_key": transfer['auth_key']}}
+        body = {"accept": {"auth_key": transfer['auth_key']}}
         req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                   fake.PROJECT_ID, transfer['id']))
 
@@ -472,6 +496,9 @@ class VolumeTransferAPITestCase(test.TestCase):
                          '2G has been consumed.',
                          res_dict['overLimit']['message'])
 
+        db.transfer_destroy(context.get_admin_context(), transfer['id'])
+        db.volume_destroy(context.get_admin_context(), volume_id)
+
     def test_accept_transfer_with_VolumeLimitExceeded(self):
 
         def fake_transfer_api_accept_throwing_VolumeLimitExceeded(cls,
@@ -486,8 +513,7 @@ class VolumeTransferAPITestCase(test.TestCase):
         volume_id = self._create_volume()
         transfer = self._create_transfer(volume_id)
 
-        body = {"accept": {"id": transfer['id'],
-                           "auth_key": transfer['auth_key']}}
+        body = {"accept": {"auth_key": transfer['auth_key']}}
         req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
                                   fake.PROJECT_ID, transfer['id']))
 
@@ -503,3 +529,28 @@ class VolumeTransferAPITestCase(test.TestCase):
         self.assertEqual("VolumeLimitExceeded: Maximum number of volumes "
                          "allowed (1) exceeded for quota 'volumes'.",
                          res_dict['overLimit']['message'])
+
+        db.transfer_destroy(context.get_admin_context(), transfer['id'])
+        db.volume_destroy(context.get_admin_context(), volume_id)
+
+    def test_accept_transfer_with_auth_key_null(self):
+        volume_id = self._create_volume(size=5)
+        transfer = self._create_transfer(volume_id)
+        body = {"accept": {"auth_key": None}}
+
+        req = webob.Request.blank('/v2/%s/os-volume-transfer/%s/accept' % (
+                                  fake.PROJECT_ID, transfer['id']))
+        req.body = jsonutils.dump_as_bytes(body)
+        req.method = 'POST'
+        req.headers['Content-Type'] = 'application/json'
+        req.headers['Accept'] = 'application/json'
+        res = req.get_response(fakes.wsgi_app(
+            fake_auth_context=self.user_ctxt))
+        res_dict = jsonutils.loads(res.body)
+
+        self.assertEqual(http_client.BAD_REQUEST,
+                         res_dict['badRequest']['code'])
+        self.assertEqual(http_client.BAD_REQUEST, res.status_int)
+
+        db.transfer_destroy(context.get_admin_context(), transfer['id'])
+        db.volume_destroy(context.get_admin_context(), volume_id)
