@@ -16,6 +16,7 @@ import mock
 import taskflow.engines
 
 from cinder import context
+from cinder import exception
 from cinder import test
 from cinder.tests.unit import fake_constants as fakes
 from cinder.tests.unit import fake_volume
@@ -24,6 +25,11 @@ from cinder.volume.flows.api import manage_existing
 from cinder.volume.flows import common as flow_common
 from cinder.volume.flows.manager import manage_existing as manager
 
+if hasattr(inspect, 'getfullargspec'):
+    getargspec = inspect.getfullargspec
+else:
+    getargspec = inspect.getargspec
+
 
 class ManageVolumeFlowTestCase(test.TestCase):
 
@@ -33,7 +39,7 @@ class ManageVolumeFlowTestCase(test.TestCase):
         self.counter = float(0)
 
     def test_cast_manage_existing(self):
-        volume = fake_volume.fake_volume_type_obj(self.ctxt)
+        volume = fake_volume.fake_volume_obj(self.ctxt)
 
         spec = {
             'name': 'name',
@@ -57,6 +63,28 @@ class ManageVolumeFlowTestCase(test.TestCase):
         create_what.pop('volume_id')
         task.execute(self.ctxt, **create_what)
 
+    def test_create_db_entry_task_with_multiattach(self):
+
+        fake_volume_type = fake_volume.fake_volume_type_obj(
+            self.ctxt, extra_specs={'multiattach': '<is> True'})
+
+        spec = {
+            'name': 'name',
+            'description': 'description',
+            'host': 'host',
+            'ref': 'ref',
+            'volume_type': fake_volume_type,
+            'metadata': {},
+            'availability_zone': 'availability_zone',
+            'bootable': 'bootable',
+            'volume_type_id': fake_volume_type.id,
+            'cluster_name': 'fake_cluster'
+        }
+        task = manage_existing.EntryCreateTask(fake_volume_api.FakeDb())
+
+        result = task.execute(self.ctxt, **spec)
+        self.assertTrue(result['volume_properties']['multiattach'])
+
     @staticmethod
     def _stub_volume_object_get(self):
         volume = {
@@ -96,7 +124,57 @@ class ManageVolumeFlowTestCase(test.TestCase):
         task = manager.PrepareForQuotaReservationTask(mock_db, mock_driver)
 
         task.revert(self.ctxt, mock_result, mock_flow_failures, volume_ref)
-        mock_error_out.assert_called_once_with(volume_ref, reason=mock.ANY)
+        mock_error_out.assert_called_once_with(volume_ref,
+                                               reason='Volume manage failed.',
+                                               status='error_managing')
+
+    def test_prepare_for_quota_reservation_with_wrong_volume(self):
+        """Test the class PrepareForQuotaReservationTas with wrong vol."""
+        mock_db = mock.MagicMock()
+        mock_driver = mock.MagicMock()
+        wrong_volume = mock.MagicMock()
+        mock_manage_existing_ref = mock.MagicMock()
+        mock_except = exception.CinderException
+
+        mock_driver.manage_existing_get_size.side_effect = mock_except
+        task = manager.PrepareForQuotaReservationTask(mock_db, mock_driver)
+        self.assertRaises(exception.CinderException,
+                          task.execute,
+                          self.ctxt,
+                          wrong_volume,
+                          mock_manage_existing_ref)
+
+    def test_manage_existing_task(self):
+        """Test the class ManageExistingTask."""
+        mock_db = mock.MagicMock()
+        mock_driver = mock.MagicMock()
+        mock_volume = mock.MagicMock()
+        mock_manage_existing_ref = mock.MagicMock()
+        mock_size = mock.MagicMock()
+
+        task = manager.ManageExistingTask(mock_db, mock_driver)
+        rv = task.execute(self.ctxt, mock_volume, mock_manage_existing_ref,
+                          mock_size)
+
+        expected_output = {'volume': mock_volume}
+        self.assertDictEqual(rv, expected_output)
+
+    def test_manage_existing_task_with_wrong_volume(self):
+        """Test the class ManageExistingTask with wrong volume."""
+        mock_db = mock.MagicMock()
+        mock_driver = mock.MagicMock()
+        mock_volume = mock.MagicMock()
+        mock_volume.update.side_effect = exception.CinderException
+        mock_manage_existing_ref = mock.MagicMock()
+        mock_size = mock.MagicMock()
+
+        task = manager.ManageExistingTask(mock_db, mock_driver)
+        self.assertRaises(exception.CinderException,
+                          task.execute,
+                          self.ctxt,
+                          mock_volume,
+                          mock_manage_existing_ref,
+                          mock_size)
 
     def test_get_flow(self):
         mock_volume_flow = mock.Mock()
@@ -107,7 +185,9 @@ class ManageVolumeFlowTestCase(test.TestCase):
             'context': mock.sentinel.context,
             'volume': mock.sentinel.volume,
             'manage_existing_ref': mock.sentinel.ref,
-            'optional_args': {'is_quota_committed': False},
+            'group_snapshot': None,
+            'optional_args': {'is_quota_committed': False,
+                              'update_size': True}
         }
 
         manager.get_flow(
@@ -133,6 +213,7 @@ class ManageVolumeFlowTestCase(test.TestCase):
             'context',
             'volume',
             'manage_existing_ref',
+            'group_snapshot',
             'optional_args',
         ]
 
@@ -146,10 +227,10 @@ class ManageVolumeFlowTestCase(test.TestCase):
                 for p in task.default_provides:
                     provides.add(p)
 
-            execute_args = inspect.getargspec(task.execute)[0]
+            execute_args = getargspec(task.execute)[0]
             execute_args = [x for x in execute_args if x not in provides]
             [self.assertIn(arg, param_names) for arg in execute_args]
 
-            revert_args = inspect.getargspec(task.revert)[0]
+            revert_args = getargspec(task.revert)[0]
             revert_args = [x for x in revert_args if x not in revert_provides]
             [self.assertIn(arg, param_names) for arg in revert_args]
